/*

Matrix Digital Rain implementation for the ZX Spectrum

Coded for the Matrix Challenge by Marco Varesio (Marco's Retrobits) in 2020 
https://retrobits.itch.io/
https://retrobits.altervista.org/
https://somebitsofme.altervista.org/

While the code rain animation is running, you can change the effects by pressing the following numeric keys:
7: set mode to "inverted drops"
8: set mode to "normal"
9: set mode to "woman in red"

For each effect, you can change the colour by pressing the following numeric keys:
1: blue
2: red
3: magenta 
4: green
5: cyan
6: yellow


Compile and run with:

zcc +zx -vn -startup=1 -clib=sdcc_iy -lm woman.c udg.c coderain.c -o coderain -create-app
C:\marco\emu\spectrum\emu\fuse-1.5.7-win32\fuse coderain.tap

*/

#pragma output CLIB_MALLOC_HEAP_SIZE = 0
#pragma output CRT_ORG_CODE = 0x8184   // move org just above the im2 table and jump (see below)

#include <arch/zx.h>
#include <im2.h>
#include <input.h>
#include <intrinsic.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stropts.h>
#include <z80.h>

#define ROWS 24
#define COLS 32

// Woman in red picture
extern unsigned char data[];

extern unsigned char udg[];

// Background backup
unsigned char back[6144];

typedef struct
{
	float pos;
	float speed;
	int length;
	int maxLength;
} columnStruct;

columnStruct columnsStatus[COLS];

uint16_t frames = 0;

#define MODE_NORMAL		0
#define MODE_INV_DROP	1
#define MODE_WOMAN		2

uint8_t mode = MODE_NORMAL;
uint8_t colour = INK_GREEN;

IM2_DEFINE_ISR_8080(isr)
{
	frames++;
}

void randomChar()
{
	uint8_t c = rand()%COLS;
	uint8_t r = rand()%ROWS;
	uint16_t addr = 22528 + c + (32 * r);
	uint8_t attr = *((uint8_t *)addr);
	if (attr != 0)
	{
		uint8_t ink = attr & 7;
		uint8_t paper = (attr >> 3) & 7;
		uint8_t bright = (attr >> 6) & 1;

		// Paper
		printf("\x11%c", 0x30 + paper);
		// Ink
		printf("\x10%c", 0x30 + ink);
		// Bright
		printf("\x13%c", 0x30 + bright);

		void *old_font;
		
		int ch = rand()%31 + 57;
		
		if (ch > 64)
			old_font = (void *)ioctl(1, IOCTL_OTERM_FONT, udg - 'A'*8);
		
		printf("\x16%c%c%c", c+1, r+1, ch);
		
		if (ch > 64)
			ioctl(1, IOCTL_OTERM_FONT, old_font);
	}
}

void init()
{
	// UDGs
	// startup = 1 does not support UDGs. See: https://www.z88dk.org/forum/viewtopic.php?pid=15311
	for (uint8_t r=1; r<=ROWS; r++)
	{
		for (uint8_t c=1; c<=COLS; c++)
		{
			void *old_font;
			
			int ch = rand()%31 + 57;
			
			if (ch > 64)
				old_font = (void*)ioctl(1, IOCTL_OTERM_FONT, udg - 'A'*8);
			
			printf("\x16%c%c%c", c, r, ch);
			
			if (ch > 64)
				ioctl(1, IOCTL_OTERM_FONT, old_font);

		}
	}	

	/*
	// Standard chars
	for (uint8_t r=1; r<=ROWS; r++)
	{
		for (uint8_t c=1; c<=COLS; c++)
		{
			printf("\x16%c%c", c, r);
			putchar(rand()%95 + 33);
		}
	}
	*/
}

void initWoman()
{
	zx_cls_attr(PAPER_RED | INK_RED | BRIGHT);
	memcpy((void *)0x4000, data, 6144);
}

void initDrop(columnStruct *col)
{
	int rnd = rand();
	col->length = 1;
	col->maxLength = (rnd & 7) + 18;
	col->speed = 0.85f;
	if (rnd > 28762)
		col->speed = 0.6f;
	else if (rnd > 24756)
		col->speed = 0.475f;
	col->pos = 16 - (rnd & 31);
}

void main()
{
	im2_init((void*)0x8000);           // set z80's I register to point at 0x8000, set im2 mode
	memset((void*)0x8000, 0x81, 257);  // make a 257-byte im2 table at 0x8000 containg 0x81 bytes
	// On interrupt the z80 will jump to address 0x8181	
	z80_bpoke(0x8181, 0xc3);                // z80 JP instruction
	z80_wpoke(0x8182, (unsigned int)isr);   // to the isr routine
	// Now enable interrupts
	intrinsic_ei();  // inlining ei would disturb the optimizer
	
	memset(columnsStatus, 0, COLS * sizeof(columnStruct));
	
	zx_border(INK_BLACK);
	zx_cls(PAPER_BLACK | INK_BLACK);
	
	// PAPER BLACK
	puts("\x11\x30");

	// INK GREEN; AT COL 5, ROW 21; BRIGHT 1
	puts("\x10\x34\x16\x06\x16\x13\x31The Speccy has you...");

	// INK BLACK
	puts("\x10\x30");

	init();
	
	for (uint8_t c=0; c<COLS; c++)
	{
		columnStruct *col = &(columnsStatus[c]);
		initDrop(col);
	}
	
	for (;;)
	{
		uint16_t frame = frames;
		
		uint8_t ch = in_inkey();
		// 7: set mode to "inverted drops"
		if (ch == 55)
		{
			if (mode > MODE_INV_DROP)
			{
				// Restore background
				memcpy((void *)0x4000, back, 6144);
				zx_cls_attr(PAPER_BLACK | INK_BLACK);
				colour = INK_GREEN;
			}
			mode = MODE_INV_DROP;
		}
		// 8: set mode to "normal"
		else if (ch == 56)
		{
			if (mode > MODE_INV_DROP)
			{
				// Restore background
				memcpy((void *)0x4000, back, 6144);
				zx_cls_attr(PAPER_BLACK | INK_BLACK);
				colour = INK_GREEN;
			}
			mode = MODE_NORMAL;
		}
		// 9: set mode to "woman in red"
		else if (ch == 57)
		{
			if (mode != MODE_WOMAN)
			{
				mode = MODE_WOMAN;

				// Backup background
				memcpy(back, (void *)0x4000, 6144);
				colour = INK_RED;
				initWoman();
			}
		}
		// 1..6: change colour
		else if (ch > 48 && ch < 55)
		{
			if (mode == MODE_WOMAN)
			{
				if (colour != ch - 48)
				{
					colour = ch - 48;	
					zx_cls_attr((colour << 3) | colour | BRIGHT);
				}
			}
			else
			{
				colour = ch - 48;
			}
			
		}

		// Random char
		if (mode < MODE_WOMAN)
		{
			randomChar();
		}

		for (uint8_t c=0; c<32; c++)
		{
			columnStruct *col = &(columnsStatus[c]);
						
			int pos = (int)col->pos;
			
			int i = pos;
			if (i >= 0 && i < ROWS && i >= pos - col->length)
			{
				if (mode == MODE_WOMAN)
				{
					z80_bpoke(22528 + c + (32 * i), (colour << 3) | (colour < 4? INK_YELLOW : INK_WHITE) | BRIGHT);
				}
				else if (mode == MODE_NORMAL)
				{
					z80_bpoke(22528 + c + (32 * i), INK_WHITE | BRIGHT);
				}
				else
				{
					z80_bpoke(22528 + c + (32 * i), INK_BLACK | (colour << 3) | BRIGHT);	
				}
			}
			
			i = pos - 1;
			if (i >= 0 && i < ROWS && i >= pos - col->length)
			{
				if (mode == MODE_WOMAN)
					z80_bpoke(22528 + c + (32 * i), (colour << 3) | INK_BLACK | BRIGHT);
				else
					z80_bpoke(22528 + c + (32 * i), colour | BRIGHT);
			}

			i = pos - 4;
			if (i >= 0 && i < ROWS && i >= pos - col->length)
			{
				if (mode == MODE_WOMAN)
					z80_bpoke(22528 + c + (32 * i), (colour << 3) | INK_BLACK | BRIGHT);
				else
					z80_bpoke(22528 + c + (32 * i), colour);
			}

			i = pos - col->length;
			if (i >= 0 && i < ROWS && i >= pos - col->length)
			{
				if (mode == MODE_WOMAN)
					z80_bpoke(22528 + c + (32 * i), (colour << 3) | colour | BRIGHT);
				else
					z80_bpoke(22528 + c + (32 * i), INK_BLACK);
			}
			

			col->pos += col->speed;

			if (((int)(col->pos) - pos > 0) && (col->length < col->maxLength))
			{
				(col->length)++;
			}

			if (col->pos >= ROWS + col->length)
			{
				initDrop(col);
			}
		}

		while (frame == frames);
	}
}
